#!/usr/bin/env perl 
# vim:ai

# (c) Klaus Franken
# finit@klaus.franken.de
# $Id: finit.pl,v 1.5 1999/03/26 00:00:34 kfr Exp $
# $Revision: 1.5 $

$VERSION="alpha 1";

# globals:
# %GLOBAL
    # key: variable
    # val: value
# %SERVICES
    # key: service
    # val: 1 = service known
# %NEED
    # key: service
    # val: comma-seperated list of needed services
# %POSTSTART
    # key: service
    # val: comma-seperated list of post-programms
# %NOSERVICE
    # key: service
    # val: 1 = service will be started everytime it is called, 
    #          never be stopped, never be restarteded
# %RREASON
    # key: service
    # val: "startreason" of service was started
# %RSTARTIME
    # key: service
    # val: time service started
# %RSTART
    # key: service
    # val: 1 = service started
# %RUNLEVELS
    # key: runlevel
    # val: 1 = runlevel known
# %RUNLEVEL
    # key: runlevel
    # val: comma-seperated list of services


######################################################################
# Main
######################################################################

# Defaults
$GLOBAL{'varrun'}='/var/run/finit';
$GLOBAL{'varlog'}='/var/log/finit';
$GLOBAL{'scripts'}='/etc/rc.d';
$GLOBAL{'systemconf'}='/etc/finit.conf';
$GLOBAL{'localconf'}='/etc/finit.conf.local';
$GLOBAL{'backuplog'}=2;
$GLOBAL{'verbose'}=0;
$GLOBAL{'testmode'}=0;
$GLOBAL{'initRead'}=0;

&commandLine(@ARGV);
exit 0;


######################################################################
# Functions
######################################################################

sub finitInit {
    return(0) if $GLOBAL{'initRead'};

    $GLOBAL{'initRead'}=1;
    &readFile($GLOBAL{'systemconf'});
    &readFile($GLOBAL{'localconf'}, 1);
    ##&readFile($FINITD);
    &readInit($GLOBAL{'scripts'});
    &parseRun();

    $|=1; # FIXME: only usefull if output goes to screen
}

sub commandLine {
    local(@ARGV)=@_;

    # parse the commandline
    # after first parse: call finitInit (so you can change the configfile
    #                                    befor)

    local($arg, $arg2, @ARGV2);
    &usage() unless ($#ARGV >= 0);

    # first parse
    while (defined($arg=shift(@ARGV))) {
        if ($arg eq '-V') {
            print "this is finit.pl, version $VERSION\n";
            return(0);
        } elsif ($arg eq '-f') {
            $arg=shift(@ARGV);
            &usage('-f needs an argument') if (! $arg || $arg =~ /^-/);
            $GLOBAL{'systemconf'}=$arg;
        } elsif ($arg eq '--help') {
            &usage();
        } elsif ($arg eq '-v') {
            $GLOBAL{'verbose'}=1;
        } elsif ($arg eq '-test') {
            $GLOBAL{'testmode'}=1;
        } else {
            push (@ARGV2, $arg);
        }
    }
    @ARGV=@ARGV2;

    # Init
    &finitInit();

    # second parse
    while (defined($arg=shift(@ARGV))) {
        if ($arg eq '-start') {
            $arg=shift(@ARGV);
            &usage('-start needs an argument')
                if (! defined($arg) || $arg =~ /^-/);
            &startService($arg, 'manual');
        } elsif ($arg eq '-restart') {
            $arg=shift(@ARGV);
            &usage('-restart needs an argument')
                if (! defined($arg) || $arg =~ /^-/);
            &reStartService($arg);
        } elsif ($arg eq '-stop') {
            $arg=shift(@ARGV);
            &usage('-stop needs an argument')
                if (! defined($arg) || $arg =~ /^-/);
            &stopService($arg);
        } elsif ($arg eq '-STOP') {
            $arg=shift(@ARGV);
            &usage('-STOP needs an argument')
                if (! defined($arg) || $arg =~ /^-/);
            &stopService($arg,1);
        } elsif ($arg eq '-list') {
            &listServices(1);
        } elsif ($arg eq '-listall') {
            &listServices();
        } elsif ($arg eq '-listtree') {
            &listServicesTree();
        } elsif ($arg eq '-startrunlevel') {
            $arg=shift(@ARGV);
            &usage('-startrunlevel needs an argument')
                if (! defined($arg) || $arg =~ /^-/);
            &startRunlevel($arg);
        } elsif ($arg eq '-stoprunlevel') {
            $arg=shift(@ARGV);
            &usage('-stoprunlevel needs an argument')
                if (! defined($arg) || $arg =~ /^-/);
            &stopRunlevel($arg);
        } elsif ($arg eq '-gorunlevel') {
            $arg=shift(@ARGV);
	    #print "$arg:", &getRunlevel($arg), "\n";
            &usage('-gorunlevel needs an argument') 
                if (! defined($arg) || $arg =~ /^-/);
            &goRunlevel($arg);
        } elsif ($arg eq '-changerunlevel') {
            $arg=shift(@ARGV);
            $arg2=shift(@ARGV);
            &usage('-changerunlevel needs two argument')
                if (! defined($arg)||! defined($arg2)||$arg =~ /^-/||$arg2 =~ /^-/);
            &changeRunlevel($arg, $arg2);
        } else {
            &usage("unknown argument: $arg");
        }
    }
}

sub usage {
    local($unknown) = @_;
    print STDERR "$unknown\n" if $unknown;
    print STDERR "usage: $0 [-t] ...\n";
    exit 1
}

sub verbose {
    local($msg) = @_;

    return() unless $GLOBAL{'verbose'};
    print "X: $msg\n";
}

sub error {
    local($msg) = @_;

    print "E: $msg\n";
}

sub readFile {
    local($file, $mustnotbethere) = @_;

    # read a configuration file

    return(0) if (! -r $file && $mustnotbethere);
    &verbose("reading file, '$file'");
    open(IN, $file) ||
        die "can't open file '$file' for reading\n";
    $section = 'services';
    while ($line = <IN>) {
        chop($line);
        $line =~ s/#.*$//;
        while ($line) {
            $line = &truncBlanks($line);
            next unless $line;

            if ($line =~ s/^\[(.*)\]$//) {
                $section = $1;
                &verbose("new section: '$section'");
            } elsif ($line =~ s/^(\w+)://) {
                $service = $1;
                if ($section eq 'services') {
                    if (-x "$GLOBAL{'scripts'}/$service") {
                        $SERVICES{$service} = 1;
                    } else{
                        # FIXME: hmmm
                        $SERVICES{$service} = 0;
                    }
                } elsif ($section eq 'runlevels') {
                    $RUNLEVELS{$service} = 1;
                } elsif ($section eq 'global') {
                } else {
                    &verbose("ignore section '$section', block '$service' \
                        command '$cmd'");
                }
            } else {
                foreach $cmd (split(/;/, $line)) {
                    $cmd = &truncBlanks($cmd);
                    if ($section eq 'services') {
                        &saveService($service, $cmd);
                    } elsif ($section eq 'runlevels') {
                        &saveRunlevel($service, $cmd);
                    } elsif ($section eq 'global') {
                        &saveGlobal($service, $cmd);
                    } else {
                        &verbose("ignore section '$section', block '$service' \
                            command '$cmd'");
                    }
                }
                $line = '';
            }
        }
    }
    close(IN);
}

sub readInit {
    local($dir)=@_;
    opendir(S, $dir)
        || die "can't opendir $dir: $!";
    while($file = readdir(S)) {
        next if $file =~ /^\./;
        next unless -x "$dir/$file";
        next if -d "$dir/$file";
        $SERVICES{$file} = 1;
    }
    closedir(S);
    return(0);
}

sub saveService {
    local($service, $cmd)=@_;

    # save information about service, one command per call

    local($fparameter);

    $cmd =~ /(\w+)\s*(.*)/;
    $cmdname = $1;
    $parameter = $2;

    $fparameter=$parameter;
    $fparameter =~ s/\s+//g;
    @parlist = split(/,/, $fparameter);

    if ($cmdname eq 'clear') {
        #$SERVICES{$service} = 0; # FIXME: we need a 'delete' command too
        $NEED{$service} = '';
        $NOSERVICE{$service} = 0;
        $POSTSTART{$service} = '';
    } elsif ($cmdname eq 'needs') {
        foreach $par (@parlist) {
            # check for loop
            local($server, $err);
            foreach $server (split(/,/, &getServers($par, 1))) {
                if ($server eq $service) {
                    &error("LOOP detected, ignoring: '$service' needs '$par'");
                    $err=1;
                }
            }
            if (!$err) {
                &verbose("'$service' needs '$par'");
                $NEED{$service} .= "$par,";
            }
        }
    } elsif ($cmdname eq 'poststart') {
        foreach $par (@parlist) {
            #### check for loop
            local($server, $err);
            ###foreach $server (split(/,/, &getServers($par, 1))) {
                ###if ($server eq $service) {
                    ###&error("LOOP detected, ignoring: '$service' needs '$par'");
                    ###$err=1;
                ###}
            ###}
            if (!$err) {
                &verbose("'$service' poststart '$par'");
                $POSTSTART{$service} .= "$par,";
            }
        }
    } elsif ($cmdname eq 'noservice') {
        $NOSERVICE{$service} = 1;

    #
    # not yet implemented:
    #
    } elsif ($cmdname eq 'wish') {
    } elsif ($cmdname eq 'before') {
    } elsif ($cmdname eq 'provides') {
    } elsif ($cmdname eq 'once') {
    } elsif ($cmdname eq 'preprg') {
    }
}

sub saveRunlevel {
    local($runlevel, $parameter)=@_;

    # save information about a runlevel

    local($fparameter, $par, @parlist);

    $fparameter=$parameter;
    $fparameter =~ s/\s+//g;
    #@parlist = split(/,/, $fparameter);
    @parlist = &csl2L($fparameter);

    foreach $par (@parlist) {
        #print "save in Runlevel '$runlevel': '$par'\n";
        $RUNLEVEL{$runlevel} .= "$par,";
    }
}

sub saveGlobal {
    local($var, $parameter)=@_;

    # FIXME: syntax check
    $GLOBAL{$var} = $parameter;
}

sub truncBlanks {
    local($line) = @_;

    $line =~ s/^\s*//;
    $line =~ s/\s*$//;

    return($line);
}

sub startService {
    local($services, $reason, $testonly, $restart) = @_;

    # start services
    #   $services: csl of services
    #   $reason: why it was started: 'manual', 'needed_by <service>'
    #   $testonly: test if all needed service are startable
    #   $restart: Restart-Mode 'noservice' is ignored
    # Return: (only usefull if $services is exactly one service!)
    #   0: service already started
    #   1: ok
    #  -1: error

    local($service, $ok, $notok);
    $ok=0;
    $notok=0;

    SERVICE: foreach $service (&csl2L($services)) {
        if ($RSTART{$service} && (! $NOSERVICE{$service} || $restart)) {
            &verbose("service '$service' allready started");
            next;
        } else {
            if (! $SERVICES{$service}) {
                &verbose("Service '$service' unknown");
                $notok++;
                next;
            }
            local($err);
            foreach $need (&csl2L(&getServers($service, 0).",$POSTSTART{$service}")) {
                if (&startService($need, '', 1) < 0) {
                    &error("don't now how to start '$need', needed by '$service'");
                    $err++;
                }
            }
            if ($err) {
                $notok++;
                next;
            }
            if ($testonly) {
                $ok=1;
                next;
            }
            foreach $need (&csl2L(&getServers($service, 0))) {
                if (&startService($need, "needed_by $service") < 0) {
                    $notok++;
                    next SERVICE;
                }
            }
            # start $service
            print "starting '$service' ...";
            &backupFile("$GLOBAL{'varlog'}/$service.start");
            $startcode = &startCmd("$GLOBAL{'scripts'}/$service start", "$GLOBAL{'varlog'}/$service.start");
            if ($startcode) {
                print " ERROR: $startcode\n";
                ### FIXME print "$starttext";
                $RSTART{$service}=0;
                $notok++;
            } else {
                print " ok\n";
                $date=`date`; chop($date);
                $RSTARTTIME{$service}=$date;
                $RREASON{$service}=$reason;
                $RSTART{$service}=1;
                $ok++;
                # POSTSTART ?
                foreach $need ($POSTSTART{$service}) {
                    if (&startService($need, "poststart from $service") < 0) {
                        &error("can't start poststart '$need' for '$service', ignoring");
                    }
                }
            }
            &storeRun($service);
        }
    }
    if ($notok) {
        return(-1);
    } elsif ($ok) {
        return(1);
    } else {
        return(0);
    }
}

sub stopService {
    local($service, $all)=@_;

    # stop Service $service
    # $all: 1: stop all services who need me too (recursiv)

    # started?
    if (! $RSTART{$service}) {
        &verbose("'$service' not started, ok");
        return(0);
    }
    # knwon service?
    if (! $SERVICES{$service}) {
        &verbose("'$service' not known");
        return(0);
    }

    # somebody love^H^H^H^H^H need me?
    $clients = &getClients($service, 0, 1);
    if ($clients) {
        if ($all) {
            #foreach $client (split(/,/, $clients)) {
            foreach $client (&csl2L($clients)) {
                &stopService($client, $all);
            }
        } else {
            &error("can't stop '$service', clients: $clients");
            return(-1);
        }
    }
    # stop
    #&verbose("stopping service '$service'");
    print "stopping '$service' ...";
    &backupFile("$GLOBAL{'varlog'}/$service.stop");
    $startcode = &startCmd("$GLOBAL{'scripts'}/$service stop", "$GLOBAL{'varlog'}/$service.stop");
    if ($startcode) {
        print " ERROR: $startcode\n";
        ### FIXME print "$starttext";
        $RSTART{$service}=1; # FIXME: are we started or not?
        $return=-1;
    } else {
        print " ok\n";
        $RSTART{$service}=0;
        $return=1;
    }
    &storeRun($service);
    return($return);
}

sub stopServicesNoClients {
    
    # stop all services without clients

    local($stopped, $service);
    while(1) {
        $stopped=0;
        foreach $service (&csl2L(&getStartedList())) {
            if (! &getClients($service, 0, 1)) {
                if (&stopService($service) > 0) {
                    $stopped++;
                }
            }
        }
        print "stopped: $stopped\n";
        last unless $stopped;
    }
}

sub reStartService {
    local($service) = @_;

    # stop service $service and all clients, then restart all currently
    #   started services

    if ($RSTART{$service}) {
        local($started);
        $started = &getStartedList();
        &stopService($service, 1);
        &startService($started, "restart by $service", 0, 1);
    } else {
        &startService($service);
    }
}

sub getClientsAss {
    local($service, $all, *CLIENTS)=@_;

    # get array (%CLIENTS) of clients of $service
    #   all: =1: all clients (recurciv)

    local($tservice);

    foreach $tservice (keys(%SERVICES)) {
        foreach $need (split(/,/, $NEED{$tservice})) {
            if ($service eq $need) {
                $CLIENTS{$tservice}=1;
                if ($all) {
                    &getClientsAss($tservice, $all, *CLIENTS);
                }
            }
        }
    }
}

sub getClients {
    local($service, $all, $startonly)=@_;

    # get a comma-seperated list of clients of $service
    #   all: =1: all clients (recurciv)
    #   $startonly==1: only started services

    local($ret, $tservice, %CLIENTS);

    &getClientsAss($service, $all, *CLIENTS);
    foreach $tservice (sort(keys(%CLIENTS))) {
        if (!$startonly || $RSTART{$tservice}) {
            $ret .= "$tservice,";
        }
    }
    chop($ret);
    return($ret);
}

sub getServersAss {
    local($service, $all, *SERVERS)=@_;

    # get array (%SERVERS) of servers of $service
    #   all: =1: all servers (recurciv)

    local($need);

    foreach $need (split(/,/, $NEED{$service})) {
        $SERVERS{$need}=1;
        if ($all) {
            &getServersAss($need, $all, *SERVERS);
        }
    }
}

sub getServers {
    local($service, $all)=@_;

    # get a comma-seperated list of servers of $service
    #   all: =1: all servers (recurciv)

    local($ret);
    if (! $all) {
        $ret=$NEED{$service};
    } else {
        local($need, %SERVERS);
        &getServersAss($service, $all, *SERVERS);
        foreach $need (sort(keys(%SERVERS))) {
            $ret .= "$need,";
        }
    }
    chop($ret);
    return($ret);
}

sub listServices {
    local($startonly, $service)=@_;

    # list all services
    #   $startonly==1: only started services
    #   $service: print this service or all if empty        

    local($val, $list);
    
    if (! $service) {
        if ($startonly) {
            $list=&getStartedList();
        } else {
            $list=&getServiceList();
        }
        foreach $service (sort(split(/,/, $list))) {
            &listServices(0, $service);
        }
    } else {
        print "No-" if $NOSERVICE{$service};
        print "Service:", &markServicename($service), "\n";
        ($val=$RSTARTTIME{$service}) &&
            print "\tStarttime: $val\n";
        ($val=$RREASON{$service}) &&
            print "\tStartreason: $val\n";
        ($val=&markServicename(&getServers($service))) &&
            print "\tServers: $val\n";
        ($val=&markServicename(&getClients($service))) &&
            print "\tClients: $val\n";
    }
}

sub listServicesTree {
    local($startonly, $service, $insert)=@_;
    
    # print a "tree" of service-relations
    #   $startonly==1: only started services
    #   $service: print this service or all if empty        
    #   $insert: 

    if (! $service) {
        local($list, %PRINTED);
        if ($startonly) {
            $list=&getStartedList();
        } else {
            $list=&getServiceList();
        }
        foreach $service (split(/,/, $list)) {
            foreach $need (split(/,/, &getServers($service, 0))) {
                $PRINTED{$need} = 1;
            }
        }
        foreach $service (sort(split(/,/, $list))) {
            next if $PRINTED{$service};
            &listServicesTree(0, $service, $insert);
        }
    } else {
        for ($tab=1; $tab<=$insert; $tab++) {
            print "\t";
        }
        print &markServicename($service), ":\n";
        foreach $need (sort(split(/,/, &getServers($service, 0)))) {
            &listServicesTree(0, $need, $insert+1);
        }
    }
}

sub markServicename {
    local($servicelist)=@_;

    # mark every servicename as startet, unknown ...
    # $servicelist: comma-seperated list of servicenames

    local($service, $ret, $mark);
    foreach $service (split(/,/, $servicelist)) {
        $mark='';
        if (! $SERVICES{$service}) {
            $mark .= '!';
        }
        if (! $RSTART{$service}) {
            $mark .= '-';
        }
        $ret .= "$service";
        if ($mark) {
            $ret .= "($mark)";
        }
        $ret .= ',';
    }
    chop($ret);
    return($ret);
}

sub getServiceList {

    # get a comma-seperated list of all services

    local($ret, $service);
    foreach $service (sort(keys(%SERVICES))) {
        $ret .= "$service,";
    }
    chop($ret);
    return($ret);
}

sub getStartedList {
    
    # get a comma-seperated list of started services

    local($ret, $service);

    foreach $service (sort(keys(%RSTART))) {
        if ($RSTART{$service}) {
            $ret .= "$service,";
        }
    }
    chop($ret);
    return($ret);
}

sub getStartedListByDir {
    
    # get a comma-seperated list of started services by reading $GLOBAL{'varrun'}

    local($ret);

    opendir(S, $GLOBAL{'varrun'})
        || die "can't opendir $GLOBAL{'varrun'}: $!";
    while($file = readdir(S)) {
        next if $file =~ /^\./;
        $ret .= "$file,";
    }
    chop($ret);
    closedir(S);
    #&verbose("started: $ret");
    return($ret);

}

# parseRun/storeRun
# Verfahren fuer $GLOBAL{'varrun'}/$service:
# 1. alle $GLOBAL{'varrun'}/$service sind immer aktuell im Speicher, also
#   - werden bei Programmstart mit &parseRun() eingelesen
#   - werden sofort nach Erstellen/Aenderung mit &storeRun() weggeschrieben
#     (Ausnahme: 'testmode': $RSTART wird gesetz, aber nichts gespeichert)
# 2. Kennzeichen fuer gestarteten Service '$RSTART{$service}==1'

sub parseRun {
    local($service)=@_;
    
    # parse the var/run-File for $service
    #   $service='': parse all

    if (!$service) {
        #&verbose("parse all");
        foreach $service (split(/,/, &getStartedListByDir())) {
            #&verbose("parse '$service'");
            &parseRun($service);
        }
    } else {
        if (! -r "$GLOBAL{'varrun'}/$service") {
            return(-1);
        }
        &verbose("reading '$GLOBAL{'varrun'}/$service'");
        open(R, "$GLOBAL{'varrun'}/$service")
            || die "can't open '$GLOBAL{'varrun'}/$service' for reading: $!";
        $RSTART{$service}=1;
        while($line = <R>) {
            chop($line);
            #print "line: $line\n";
            $line =~ /(\w+):\s+(.*)$/;
            $cmd=$1;
            $val=$2;
            if ($cmd eq 'starttime') {
                $RSTARTTIME{$service}=$val;
            } elsif ($cmd eq 'reason') {
                $RREASON{$service}=$val
            } else {
                #FIXME: ignoring?
                &verbose("unknown line '$line' in '$GLOBAL{'varrun'}/$service'");
            }
        }
    }
}

sub storeRun {
    local($service)=@_;

    # stores var/run-File for $service 
    #   $service='': store all

    # do nothing if we are in testmode
    return(0) if $GLOBAL{'testmode'};

    if (!$service) {
        foreach $service (split(/,/, &getStartedList())) {
            &storeRun($service);
        }
    } else {
        if ($RSTART{$service}) {
            open (R, ">>$GLOBAL{'varrun'}/$service")
                || die "can't write to '$GLOBAL{'varrun'}/$service': $!";

            print R "starttime: $RSTARTTIME{$service}\n";
            print R "reason: $RREASON{$service}\n";

            close(R);
        } else {
            unlink("$GLOBAL{'varrun'}/$service");
        }
    }
}

sub getRunlevel {
    local($runlevel, $rectest)=@_;

    # get all services for $runlevel

    local($service, $ret, %PLUSSERV, %MINUSSERV);

    # loop detection
    foreach $service (split(/,/, $rectest)) {
        if ($service eq $runlevel) {
            &error("loop detected in runlevel: '$runlevel', ignored");
            return('');
        }
    }

    foreach $service (split(/,/, $RUNLEVEL{$runlevel})) {
        if ($service =~ /-(\w+)/) {
            $service = $1;
            $MINUSSERV{$service}++;
        } elsif ($service =~ /-\((\w+)\)/) {
            $service = $1;
            foreach $service (split(/,/, 
                &getRunlevel($service, "$rectest$runlevel,"))) {
                $MINUSSERV{$service}++;
            }
        } elsif ($service =~ /\((\w+)\)/) {
            $service = $1;
            $rectest .= "$runlevel,";
            foreach $service (split(/,/, 
                &getRunlevel($service, $rectest))) {
                $PLUSSERV{$service}++;
            }
        } else {
            $PLUSSERV{$service}++;
        }
    }

    foreach $service (sort(keys(%PLUSSERV))) {
        if (($PLUSSERV{$service}-$MINUSSERV{$service}) > 0) {
            $ret .= "$service,";
        }
    }
    chop($ret);
    return($ret);
}

sub startRunlevel {
    local($runlevel)=@_;

    # start all services for runlevel $runlevel

    local($service);
    foreach $service (split/,/, &getRunlevel($runlevel)){
        &startService($service, "runlevel $runlevel");
    }
}

sub stopRunlevel {
    local($runlevel)=@_;

    # stop all services for runlevel $runlevel

    local($service);
    foreach $service (split/,/, &getRunlevel($runlevel)){
        &stopService($service);
    }
}

sub goRunlevel {
    local($runlevel)=@_;

    # start all services for runlevel $runlevel and stop all others

    local($service, $all, $tostart, $tostop);
    $tostart = &getRunlevel($runlevel);
    # expand $tostart for all services
    foreach $service (&csl2L($tostart)) {
        $all .= "$service,";
        $all .= &getServers($service,1) . ',';
    }
    #print "$tostart = $all\n";

    $tostop = &cslMinus (&getStartedList(), $all);
    foreach $service (&csl2L($tostop)) {
        unless ($NOSERVICE{$service}) {
            &stopService($service);
        }
    }
    foreach $service (&csl2L($tostart)) {
        &startService($service, "runlevel $runlevel");
    }
}

sub changeRunlevel {
    local($runlevel, $oldrunlevel)=@_;

    # start all services for runlevel $runlevel and stop all from $oldrunlevel

    local($service, $all, $allold, $tostart, $tostop);
    $tostart = &getRunlevel($runlevel);
    foreach $service (&csl2L($tostart)) {
        $all .= "$service,";
        $all .= &getServers($service,1) . ',';
    }
    $tostop = &getRunlevel($oldrunlevel);
    foreach $service (&csl2L($tostop)) {
        $allold .= "$service,";
        $allold .= &getServers($service,1) . ',';
    }

    $tostop = &cslMinus ($allold, $all);
    foreach $service (&csl2L($tostop)) {
        unless ($NOSERVICE{$service}) {
            &stopService($service);
        }
    }
    foreach $service (&csl2L($tostart)) {
        &startService($service, "runlevel $runlevel");
    }
}

sub startCmd {
    local($cmd, $logfile) = @_;

    # start a command 
    # return: exit-code from command

    local($err);
    if ($GLOBAL{'testmode'}) {
        print "Testmode: command '$cmd'; assuming OK";
        $err=0;
    } else {
        system("$cmd > $logfile 2>&1");
        $err = $?/256;
    }
    return($err);
}

sub backupFile {
    local($file)=@_;

    # backup a file with extension '.1', '.2', '.3' ...
    # maximum $GLOBAL{'backuplog'}

    local($i, $j);
    if (-r $file) {
        for ($i=1; $i < $GLOBAL{'backuplog'}; $i++) {
            $j=$i+1;
            #print "rename $file.$i $file.$j" if -r "$file.$i";
            rename("$file.$i","$file.$j") if -r "$file.$i";
        }
        rename $file,"$file.1" if $GLOBAL{'backuplog'};
    }
}


#
# csl (comma seperated lists) functions
# why csl? it's easier to store in an array.
#

sub testCsl {
    $res = &l2Csl('a', 'b', 'c', 'a');
    $req = 'a,b,c,a';
    if ($res ne $req) {
        print "required: '$req'\n";
        print "result:   '$res'\n";
        $err++;
    }

    @A = ('a');
    $res = &l2Csl(@A, &csl2L('b,c,a'));
    $req = 'a,b,c,a';
    if ($res ne $req) {
        print "required: '$req'\n";
        print "result:   '$res'\n";
        $err++;
    }

    $res = &cslAdd('a,c', 'b,c,d');
    $req = 'a,b,c,d';
    if ($res ne $req) {
        print "required: '$req'\n";
        print "result:   '$res'\n";
        $err++;
    }

    $res = &cslMinus('a,e', 'b,c,d');
    $req = 'a,e';
    if ($res ne $req) {
        print "required: '$req'\n";
        print "result:   '$res'\n";
        $err++;
    }

    # error?
    if ($err) {
        print "csl: $err error\n";
        exit($err);
    } else {
        print "csl ok\n";
    }
}

sub cslAdd {
    local($csl1, $csl2)=@_;

    # add 2 csl's - unique

    local(%NEW);

    foreach $item (&csl2L($csl1), &csl2L($csl2)) {
        $NEW{$item}=1;
    }
    return(&l2Csl(sort(keys(%NEW))));
}

sub cslMinus {
    local($csl1, $csl2)=@_;

    # $cls1 minus $csl2 - unique

    local(%ONE, %TWO);
    foreach $item (&csl2L($csl2)) {
        $TWO{$item}=1;
    }
    foreach $item (&csl2L($csl1)) {
        $ONE{$item}=1 unless $TWO{$item};
    }
    return(&l2Csl(sort(keys(%ONE))));
}

sub l2Csl {
    local(@PAR) = @_;

    # list to csl

    local($ret, $p);

    foreach $p (@PAR) {
        $ret .= "$p,";
    }
    chop($ret);
    return($ret);
}

sub csl2L {
    local($csl) = @_;

    # csl to list

    local($item, @RET);

    foreach $item (split(/,/, $csl)) {
        push(@RET, $item);
    }
    return(@RET);
}

# test
#&testCsl();
